/* dc_gui2 - a GTK+2 GUI for DCTC
 * Copyright (C) 2002 Eric Prevoteau
 *
 * global_user.c: Copyright (C) Eric Prevoteau <www@a2pb.gotdns.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/*
$Id: global_user.c,v 1.11 2003/12/06 06:53:10 ericprev Exp $
*/

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <gtk/gtk.h>

#include "global_user.h"
#include "main.h"

static GPtrArray *glob_user=NULL;

static int compare_glob_users(const void *a, const void *b)
{
	return strcmp((*((GLOB_USER **)a))->unfmt_nick,(*((GLOB_USER **)b))->unfmt_nick);
}

/******************************************/
/* sort the glob_user array by unfmt_nick */
/******************************************/
static void sort_glob_user(void)
{
	if((glob_user!=NULL)&&(glob_user->len>1))
	{
		qsort(glob_user->pdata,glob_user->len,sizeof(gpointer),compare_glob_users);
	}
}

/**********************************/
/* return the structure of a user */
/**********************************/
GLOB_USER *gu_get_user(const char *nickname)
{
	return gu_add_user(nickname,FALSE,NULL);
}

/***********************************************************/
/* return the structure of a user or create it if required */
/*********************************************************************************/
/* input: nickname to find/add                                                   */
/*        if with_creation is TRUE, when a nick is not found, it is created      */
/* output: GLOB_USER (always !=NULL)                                             */
/*         if was_created is not NULL, it is set to TRUE if the user was created */
/*********************************************************************************/
GLOB_USER *gu_add_user(const char *unfmt_nick, gboolean with_creation, gboolean *was_created)
{
	if(glob_user!=NULL)
	{
		int i;

		for(i=0;i<glob_user->len;i++)
		{
			GLOB_USER *gu;

			gu=g_ptr_array_index(glob_user,i);
			if(!strcmp(unfmt_nick,gu->unfmt_nick))
			{
				if(was_created!=NULL)
					*was_created=FALSE;
				return gu;
			}
		}
	}

	if(with_creation==FALSE)
	{
		if(was_created!=NULL)
			*was_created=FALSE;
		return NULL;			/* no user and no creation allowed */
	}
	else
	{
		GLOB_USER *ngu;

		if(glob_user==NULL)
		{
			glob_user=g_ptr_array_new();
		}
		ngu=g_malloc(sizeof(GLOB_USER));

		if(was_created!=NULL)
			*was_created=TRUE;

		ngu->unfmt_nick=g_strdup(unfmt_nick);
		ngu->unfmt_email=NULL;
		ngu->unfmt_desc=NULL;
		ngu->unfmt_size=0;
		ngu->is_op=FALSE;
		ngu->cnx_type="";
		ngu->flags=0;
		ngu->is_connected=FALSE;

		if(utf8_mode==TRUE)
		{
			ngu->fmt_nick=g_strdup(ngu->unfmt_nick);
		}
		else
		{
			ngu->fmt_nick=g_locale_to_utf8(ngu->unfmt_nick,-1,NULL,NULL,NULL);
		}
		ngu->fmt_email=NULL;
		ngu->fmt_desc=NULL;
		ngu->fmt_size=NULL;
		ngu->front_color=NULL;
		ngu->back_color=NULL;
		ngu->fast_color=NULL;
		ngu->back_present_color=NULL;

		ngu->gu_ref=g_array_new(FALSE,FALSE,sizeof(GLOB_USER_REF));
		ngu->dirty=TRUE;

		g_ptr_array_add(glob_user,ngu);
		sort_glob_user();
		return ngu;
	}
}

/*********************************************************/
/* modify a GLOB_USER entry with possible display update */
/*********************************************************/

/* -------------------------------------------------------------------------- */
/* -------------------------------------------------------------------------- */
/* -------------------------------------------------------------------------- */
gboolean gu_set_nick(GLOB_USER *gu, const char *unfmt_nick, gboolean with_display_update)
{
	if(gu->unfmt_nick)
	{
		if(!strcmp(gu->unfmt_nick,unfmt_nick))
			return FALSE;		/* no change */
		g_free(gu->unfmt_nick);
	}

	gu->unfmt_nick=g_strdup(unfmt_nick);

	if(gu->fmt_nick)
		g_free(gu->fmt_nick);
	if(utf8_mode==TRUE)
	{
		gu->fmt_nick=g_strdup(gu->unfmt_nick);
	}
	else
	{
		gu->fmt_nick=g_locale_to_utf8(gu->unfmt_nick,-1,NULL,NULL,NULL);
	}
	gu->dirty=TRUE;

	if(with_display_update)
		gu_do_user_row_update(gu);

	sort_glob_user();		/* the array is sorted on the nickname, we must keep it sorted */
	return TRUE;
}

/* -------------------------------------------------------------------------- */
/* -------------------------------------------------------------------------- */
/* -------------------------------------------------------------------------- */
gboolean gu_set_email(GLOB_USER *gu, const char *unfmt_email, gboolean with_display_update)
{
	if(gu->unfmt_email)
	{
		if(!strcmp(gu->unfmt_email,unfmt_email))
			return FALSE;		/* no change */
		g_free(gu->unfmt_email);
	}
	gu->unfmt_email=g_strdup(unfmt_email);

	if(gu->fmt_email)
		g_free(gu->fmt_email);
	if(utf8_mode==TRUE)
	{
		gu->fmt_email=g_strdup(gu->unfmt_email);
	}
	else
	{
		gu->fmt_email=g_locale_to_utf8(gu->unfmt_email,-1,NULL,NULL,NULL);
	}
	gu->dirty=TRUE;

	if(with_display_update)
		gu_do_user_row_update(gu);
	return TRUE;
}

/* -------------------------------------------------------------------------- */
/* -------------------------------------------------------------------------- */
/* -------------------------------------------------------------------------- */
gboolean gu_set_desc(GLOB_USER *gu, const char *unfmt_desc, gboolean with_display_update)
{
	if(gu->unfmt_desc)
	{
		if(!strcmp(gu->unfmt_desc,unfmt_desc))
			return FALSE;		/* no change */
		g_free(gu->unfmt_desc);
	}
	gu->unfmt_desc=g_strdup(unfmt_desc);

	if(gu->fmt_desc)
		g_free(gu->fmt_desc);
	if(utf8_mode==TRUE)
	{
		gu->fmt_desc=g_strdup(gu->unfmt_desc);
	}
	else
	{
		gu->fmt_desc=g_locale_to_utf8(gu->unfmt_desc,-1,NULL,NULL,NULL);
	}
	gu->dirty=TRUE;

	if(with_display_update)
		gu_do_user_row_update(gu);
	return TRUE;
}

/* -------------------------------------------------------------------------- */
/* -------------------------------------------------------------------------- */
/* -------------------------------------------------------------------------- */
gboolean gu_set_size(GLOB_USER *gu, const guint64 unfmt_size, gboolean with_display_update)
{
	GString *fsize;
	double vsize;

	if(gu->unfmt_size==unfmt_size)
		return FALSE;		/* no changes ? */
	
	gu->unfmt_size=unfmt_size;

	fsize=g_string_new("");
	vsize=unfmt_size;

	if(vsize>(1024.0*1024.0*1024.0))
#ifndef NO_PRINTF_LOCALE
		g_string_sprintf(fsize,"%'.2fGB",vsize/(1024.0*1024.0*1024.0));	/* NO_PRINTF_LOCAL support added */
#else
		g_string_sprintf(fsize,"%.2fGB",vsize/(1024.0*1024.0*1024.0));		
#endif
	else if(vsize>(1024.0*1024.0))
		g_string_sprintf(fsize,"%.2fMB",vsize/(1024.0*1024.0));
	else if(vsize>(1024.0))
		g_string_sprintf(fsize,"%.2fKB",vsize/(1024.0));
	else
		g_string_sprintf(fsize,"%.2fB",vsize);

	if(gu->fmt_size==NULL)
	{
		gu->fmt_size=g_strdup(fsize->str);
		gu->dirty=TRUE;
	}
	else
	{
		if(strcmp(fsize->str,gu->fmt_size))
		{
			g_free(gu->fmt_size);
			gu->fmt_size=g_strdup(fsize->str);
			gu->dirty=TRUE;
		}
		else
			with_display_update=FALSE;		/* no visible change */
	}
	g_string_free(fsize,TRUE);

	if(with_display_update)
		gu_do_user_row_update(gu);
	return TRUE;
}

/* -------------------------------------------------------------------------- */
/* -------------------------------------------------------------------------- */
/* -------------------------------------------------------------------------- */
gboolean gu_set_op_flag(GLOB_USER *gu, const gboolean is_op, gboolean with_display_update)
{
	if(gu->is_op==is_op)
		return FALSE;		/* no change */

	gu->is_op=is_op;
	if(gu->is_op)
		gu->front_color="red";
	else
		gu->front_color=NULL;
	gu->dirty=TRUE;

	if(with_display_update)
		gu_do_user_row_update(gu);
	return TRUE;
}

/* -------------------------------------------------------------------------- */
/* -------------------------------------------------------------------------- */
/* -------------------------------------------------------------------------- */
static const char *lst_cnx_type[]=
									{
										"56Kbps",
										"33.6Kbps",
										"28.8Kbps",
										"Satellite",		/* according to DC, passive mode is required here */
										"ISDN",
										"DSL",
										"Cable",
										"LAN(T1)",
										"LAN(T3)",
										NULL
									};

static const char *get_const_cnx_type(const char *cnx_type)
{
	int i;

	i=0;
	while(lst_cnx_type[i]!=NULL)
	{
		if(!strcmp(cnx_type,lst_cnx_type[i]))
			return lst_cnx_type[i];
		i++;
	}
	return "Invalid";
}

gboolean gu_set_cnx_type(GLOB_USER *gu, const char *cnx_type, gboolean with_display_update)
{
	if((gu->cnx_type!=NULL)&&(!strcmp(gu->cnx_type,cnx_type)))
		return FALSE;

	gu->cnx_type=get_const_cnx_type(cnx_type);
	gu->dirty=TRUE;

	if(with_display_update)
		gu_do_user_row_update(gu);
	return TRUE;
}

/* -------------------------------------------------------------------------- */
/* -------------------------------------------------------------------------- */
/* -------------------------------------------------------------------------- */
gboolean gu_set_flags(GLOB_USER *gu, const guint flags, gboolean with_display_update)
{
	if(gu->flags==flags)
		return FALSE;		/* no change */

	gu->flags=flags;
	if(gu->flags&2)		/* away flag */
		gu->back_color="grey87";
	else
		gu->back_color=NULL;

	if(gu->flags&8)		/* fast flag */
		gu->fast_color="green";
	else
		gu->fast_color=gu->back_color;
	gu->dirty=TRUE;

	if(with_display_update)
		gu_do_user_row_update(gu);
	return TRUE;
}

/* -------------------------------------------------------------------------- */
/* -------------------------------------------------------------------------- */
/* -------------------------------------------------------------------------- */
gboolean gu_set_is_connected(GLOB_USER *gu, const gboolean is_connected, gboolean with_display_update)
{
	if(gu->is_connected==is_connected)
		return FALSE;		/* no change */

	gu->is_connected=is_connected;

	if(gu->is_connected==TRUE)
		gu->back_present_color="green";
	else
		gu->back_present_color=NULL;
	gu->dirty=TRUE;

	if(with_display_update)
		gu_do_user_row_update(gu);
	return TRUE;
}

/* -------------------------------------------------------------------------- */
/* -------------------------------------------------------------------------- */
/* -------------------------------------------------------------------------- */
/***************************************/
/* force a redisplay of the given user */
/***************************************/
void gu_do_user_row_update(GLOB_USER *gu)
{
	int i;

	if((gu==NULL)||(gu->gu_ref->len==0))
		return;		/* nothing to update */

	for(i=gu->gu_ref->len-1;i>=0;i--)
	{
		GLOB_USER_REF *gur;

		gur=&(g_array_index(gu->gu_ref,GLOB_USER_REF,i));

		if(gtk_tree_row_reference_valid(gur->gtrr))
		{
			GtkTreePath *gur_path;
			GtkTreeIter gur_iter;

			gur_path=gtk_tree_row_reference_get_path(gur->gtrr);
			if(gtk_tree_model_get_iter(gur->gtm,&gur_iter,gur_path))
			{
				/* notify the row changes to the tree model */
				gtk_tree_model_row_changed(gur->gtm,gur_path,&gur_iter);
			}

			gtk_tree_path_free(gur_path);
		}
		else
		{
			/* invalid reference => remove it */
			gtk_tree_row_reference_free(gur->gtrr);
			g_array_remove_index_fast(gu->gu_ref,i);
		}
	}
	gu->dirty=FALSE;
}

/* -------------------------------------------------------------------------- */
/* -------------------------------------------------------------------------- */
/* -------------------------------------------------------------------------- */
/*********************************/
/* reference a new displayed row */
/********************************************/
/* input: the GtkTreeRowReference is stolen */
/********************************************/
void gu_ref(GLOB_USER *gu, GtkTreeModel *gtm, GtkTreeRowReference *gtrr)
{
	GLOB_USER_REF ngur;

	ngur.gtm=gtm;
	ngur.gtrr=gtrr;

	g_array_append_val(gu->gu_ref,ngur);
}

/************************************************/
/* reference a new displayed row using its iter */
/************************************************/
void gu_ref_from_iter(GLOB_USER *gu, GtkTreeModel *gtm, GtkTreeIter *giter)
{
	GLOB_USER_REF ngur;
	GtkTreePath *gtp;

	gtp=gtk_tree_model_get_path(gtm,giter);

	ngur.gtm=gtm;
	ngur.gtrr=gtk_tree_row_reference_new(gtm,gtp);

	g_array_append_val(gu->gu_ref,ngur);

	gtk_tree_path_free(gtp);
}

/* -------------------------------------------------------------------------- */
/* -------------------------------------------------------------------------- */
/* -------------------------------------------------------------------------- */
/***********************************/
/* unreference a new displayed row */
/***********************************/
void gu_unref(GLOB_USER *gu, GtkTreeModel *gtm, GtkTreeRowReference *gtrr)
{
	int i;
	GtkTreePath *gtp;

	gtp=gtk_tree_row_reference_get_path(gtrr);

	for(i=0;i<gu->gu_ref->len;i++)
	{
		GLOB_USER_REF *gur;

		gur=&(g_array_index(gu->gu_ref,GLOB_USER_REF,i));
		if(gur->gtm==gtm)
		{
			GtkTreePath *gur_path;

			gur_path=gtk_tree_row_reference_get_path(gur->gtrr);
			if(!gtk_tree_path_compare(gur_path,gtp))
			{
				gtk_tree_path_free(gur_path);
				gtk_tree_row_reference_free(gur->gtrr);
				g_array_remove_index_fast(gu->gu_ref,i);
				break;
			}
			gtk_tree_path_free(gur_path);
		}
	}

	gtk_tree_path_free(gtp);
}

/* -------------------------------------------------------------------------- */
/* -------------------------------------------------------------------------- */
/* -------------------------------------------------------------------------- */
/**************************************************/
/* unreference a new displayed row using its iter */
/**************************************************/
void gu_unref_from_iter(GLOB_USER *gu, GtkTreeModel *gtm, GtkTreeIter *iter)
{
	int i;
	GtkTreePath *gtp;

	gtp=gtk_tree_model_get_path(gtm,iter);

	for(i=0;i<gu->gu_ref->len;i++)
	{
		GLOB_USER_REF *gur;

		gur=&(g_array_index(gu->gu_ref,GLOB_USER_REF,i));
		if(gur->gtm==gtm)
		{
			GtkTreePath *gur_path;

			gur_path=gtk_tree_row_reference_get_path(gur->gtrr);
			if(!gtk_tree_path_compare(gur_path,gtp))
			{
				gtk_tree_path_free(gur_path);
				gtk_tree_row_reference_free(gur->gtrr);
				g_array_remove_index_fast(gu->gu_ref,i);
				break;
			}
			gtk_tree_path_free(gur_path);
		}
	}

	gtk_tree_path_free(gtp);
}

/* -------------------------------------------------------------------------- */
/* -------------------------------------------------------------------------- */
/* -------------------------------------------------------------------------- */
/*******************************************************************/
/* search inside the given user for a reference on the given model */
/*******************************************************************/
/* output: NULL= not found else a TreePath (to free) */
/*****************************************************/
GtkTreePath *gu_user_get_path_for_model(GLOB_USER *gu, GtkTreeModel *gtm)
{
	int i;
	GtkTreePath *gtp=NULL;

	for(i=0;i<gu->gu_ref->len;i++)
	{
		GLOB_USER_REF *gur;

		gur=&(g_array_index(gu->gu_ref,GLOB_USER_REF,i));
		if(gur->gtm==gtm)
		{
			gtp=gtk_tree_row_reference_get_path(gur->gtrr);
			break;
		}
	}
	return gtp;
}

/* -------------------------------------------------------------------------- */
/* -------------------------------------------------------------------------- */
/* -------------------------------------------------------------------------- */
/*******************************************************************/
/* search inside the given user for a reference on the given model */
/*******************************************************************/
/* output: TRUE if the iter is valid and "giter" is populated */
/**************************************************************/
gboolean gu_user_get_iter_for_model(GLOB_USER *gu, GtkTreeModel *gtm, GtkTreeIter *iter)
{
	int i;
	gboolean output=FALSE;

	for(i=0;i<gu->gu_ref->len;i++)
	{
		GLOB_USER_REF *gur;

		gur=&(g_array_index(gu->gu_ref,GLOB_USER_REF,i));
		if(gur->gtm==gtm)
		{
			GtkTreePath *gtp;

			gtp=gtk_tree_row_reference_get_path(gur->gtrr);

			output=gtk_tree_model_get_iter(gtm,iter,gtp);
			gtk_tree_path_free(gtp);
			break;
		}
	}
	return output;
}

/* -------------------------------------------------------------------------- */
/* -------------------------------------------------------------------------- */
/* -------------------------------------------------------------------------- */
/*******************************************************************/
/* search inside the given user for a reference on the given model */
/***********************************************************************/
/* output: NULL= not found else a TreeRowReference (must not be freed) */
/***********************************************************************/
GtkTreeRowReference *gu_user_get_reference_for_model(GLOB_USER *gu, GtkTreeModel *gtm)
{
	int i;

	for(i=0;i<gu->gu_ref->len;i++)
	{
		GLOB_USER_REF *gur;

		gur=&(g_array_index(gu->gu_ref,GLOB_USER_REF,i));
		if(gur->gtm==gtm)
		{
			return gur->gtrr;
		}
	}
	return NULL;
}

/* -------------------------------------------------------------------------- */
/* -------------------------------------------------------------------------- */
/* -------------------------------------------------------------------------- */
/*******************************************************/
/* free allocated memory and destroy a GLOB_USER entry */
/*******************************************************/
static void gu_free_entry(GLOB_USER *gu)
{
	if(gu->unfmt_nick) g_free(gu->unfmt_nick);
	if(gu->unfmt_email) g_free(gu->unfmt_email);
	if(gu->unfmt_desc) g_free(gu->unfmt_desc);
	if(gu->fmt_nick) g_free(gu->fmt_nick);
	if(gu->fmt_email) g_free(gu->fmt_email);
	if(gu->fmt_desc) g_free(gu->fmt_desc);
	if(gu->fmt_size) g_free(gu->fmt_size);

	g_array_free(gu->gu_ref,TRUE);
}

/* -------------------------------------------------------------------------- */
/* -------------------------------------------------------------------------- */
/* -------------------------------------------------------------------------- */
/*********************************************/
/* purge GLOB_USER entries without reference */
/*********************************************/
void gu_purge(void)
{
	int i;

	if(glob_user==NULL)
		return;

	for(i=glob_user->len-1;i>=0;i--)
	{
		GLOB_USER *gu;

		gu=g_ptr_array_index(glob_user,i);

		if(gu->gu_ref->len==0)
		{
			g_ptr_array_remove_index(glob_user,i);
			gu_free_entry(gu);
		}
	}
}

/* -------------------------------------------------------------------------- */
/* -------------------------------------------------------------------------- */
/* -------------------------------------------------------------------------- */
/************************************************/
/* remove all reference to a model on all users */
/************************************************/
/* useful after/before a gtk_store_clear() */
/*******************************************/
void gu_unref_model(GtkTreeModel *gtm)
{
	int i;
	int j;

	if(glob_user==NULL)
		return;

	for(i=glob_user->len-1;i>=0;i--)
	{
		GLOB_USER *gu;

		gu=g_ptr_array_index(glob_user,i);
		for(j=gu->gu_ref->len-1;j>=0;j--)
		{
			GLOB_USER_REF *gur;

			gur=&(g_array_index(gu->gu_ref,GLOB_USER_REF,j));
			if(gur->gtm==gtm)
			{
				gtk_tree_row_reference_free(gur->gtrr);
				g_array_remove_index_fast(gu->gu_ref,j);
			}
		}

		if(gu->gu_ref->len==0)
		{
			g_ptr_array_remove_index(glob_user,i);
			gu_free_entry(gu);
		}
	}
}

/************************************************/
/* remove all reference to a model on all users */
/* and mark updated user as disconnected        */
/************************************************/
/* useful before a gtk_store_clear() */
/*************************************/
void gu_unref_model_and_disconnect(GtkTreeModel *gtm)
{
	int i;
	int j;

	if(glob_user==NULL)
		return;

	for(i=glob_user->len-1;i>=0;i--)
	{
		GLOB_USER *gu;
		gboolean modified=FALSE;

		gu=g_ptr_array_index(glob_user,i);
		for(j=gu->gu_ref->len-1;j>=0;j--)
		{
			GLOB_USER_REF *gur;

			gur=&(g_array_index(gu->gu_ref,GLOB_USER_REF,j));
			if(gur->gtm==gtm)
			{
				gtk_tree_row_reference_free(gur->gtrr);
				g_array_remove_index_fast(gu->gu_ref,j);
				modified=TRUE;
			}
		}

		if(modified==TRUE)
		{
			if(gu->gu_ref->len==0)
			{
				g_ptr_array_remove_index(glob_user,i);
				gu_free_entry(gu);
			}
			else
			{
				gu_set_is_connected(gu,FALSE,TRUE);	/* mark user as disconnected and do redisplay */
			}
		}
	}
}

/* -------------------------------------------------------------------------- */
/* -------------------------------------------------------------------------- */
/* -------------------------------------------------------------------------- */
/*********************************************/
/* clear dirty flag of all GLOB_USER entries */
/*********************************************/
void gu_clear_dirty_flags(void)
{
	int i;

	if(glob_user==NULL)
		return;

	for(i=glob_user->len-1;i>=0;i--)
	{
		((GLOB_USER*)(g_ptr_array_index(glob_user,i)))->dirty=FALSE;
	}
}

/* -------------------------------------------------------------------------- */
/* -------------------------------------------------------------------------- */
/* -------------------------------------------------------------------------- */
/************************************************************/
/* compute the number of online users and their shared size */
/************************************************************/
guint32 compute_online_nb_users_and_share_size(guint64 *share_size)
{
	guint64 shs=0;
	guint32 nb_users=0;
	int i;

	if(glob_user!=NULL)
	{
		for(i=0;i<glob_user->len;i++)
		{
			GLOB_USER *gu;

			gu=g_ptr_array_index(glob_user,i);

			if(gu->is_connected==TRUE)
			{
				nb_users++;
				shs+=gu->unfmt_size;
			}
		}
	}

	*share_size=shs;
	return nb_users;
}

